/* -*-fundamental-*-
 ******************************************************************************
 *
 * File:         ps_scan.l
 * RCS:          $Header: /home/chris/Psopt/cvs/psopt/src/ps_scan.l,v 1.6 1997/08/25 20:58:16 chris Exp $
 * Description:  
 * Author:       Christian Limpach <chris@nice.ch>
 * Created:      Thu Apr 24 00:39:05 1997
 * Modified:     Mon Aug 11 22:46:33 1997 (Christian Limpach) chris@nice.ch
 * Language:     Lex
 * Package:      N/A
 * Status:       Experimental (Do Not Distribute)
 *
 * (C) Copyright 1997, Christian Limpach, all rights reserved.
 *            
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *            
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *            
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, a copy can be obtained from this
 * program's author (send electronic mail to chris@nice.ch) or from
 * the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA
 * 02139, USA.
 *
 ******************************************************************************
 */

%C

%{
#include "psopt.h"
#include "opers.h"

/* #define SHOW_SCANNING */

  char *read_string ();
  char *read_binstring ();
  char *read_to_eol ();
  int check_com_arg ();
  
  typedef union {
    char *string;
    struct ps_op *ps_op;
  } YYSTYPE;

  YYSTYPE yylval;
%}

%%

^\%\%		{
                  char *s = read_to_eol ();
                  add_code (COMMENT, strdup (s));
#ifdef SHOW_SCANNING  
		  printf ("keep comment:  %s\n", s);
#endif
                }

\%		{
                  char *s = read_to_eol ();
#ifdef SHOW_SCANNING  
		  printf ("comment:       %s\n", s);
#endif
		}

\{		{
#ifdef SHOW_SCANNING  
                  puts ("open brace:    {");
#endif
                  add_code (OPEN_BRACE, NULL);
                }
\}		{
#ifdef SHOW_SCANNING  
		  puts ("close brace:   }");
#endif
                  add_code (CLOSE_BRACE, NULL);
		}

\[		{
#ifdef SHOW_SCANNING  
                  puts ("open bracket:  [");
#endif
                  add_code (OPEN_BRACKET, NULL);
                }
\]		{
#ifdef SHOW_SCANNING  
		  puts ("close bracket: ]");
#endif
                  add_code (CLOSE_BRACKET, NULL);
		}

\(		{
                  char *s = read_string ();
                  add_code (STRING, strdup (s));
#ifdef SHOW_SCANNING  
		  printf ("string:        %s\n", s);
#endif
                }
\)		{
		  fprintf (stderr, "FEHLER...\n");
		  return (0);
#ifdef SHOW_SCANNING                  
		  puts ("close paren:   )");
#endif
		}

\<		{
                  char *s = read_binstring ();
                  add_code (BINSTRING, strdup (s));
#ifdef SHOW_SCANNING  
		  printf ("binstring:     %s\n", s);
#endif
                }
\>		{
		  fprintf (stderr, "FEHLER...\n");
		  return (0);
#ifdef SHOW_SCANNING                  
		  puts ("close cmp:     >");
#endif
		}

\/[^\][{}()<>/%\ \t\n]+	{
                          Identifier *i;
                          char *s = &yytext[1];
                          i = count_identifier (s);
                          add_code (DIRECT, i);
#ifdef SHOW_SCANNING                          
			  printf ("direct:        %s\n", yytext);
#endif
			}

[^\][{}()<>/%\ \t\n]+	{
			  if (check_com_arg () == COMMAND)
                            {
                              Identifier *i;
                              i = count_identifier (yylval.ps_op->name);
                              add_code (COMMAND, i);
#ifdef SHOW_SCANNING                              
                              printf ("command:       %s %d %d\n",
                                      yylval.ps_op->name,
                                      yylval.ps_op->in,
                                      yylval.ps_op->out);
#endif
                            }
                          else
                            {
                              Identifier *i;
                              i = count_identifier (yytext);
                              add_code (ARGUMENT, i);
#ifdef SHOW_SCANNING                              
                              printf ("argument:      %s\n", yytext);
#endif
                            }
			}

\n|\ |\t		{ /* putchar (' '); */
                        }
                        
.		ECHO ;

%%

char *
read_string ()
{
  char c;
  static char *s;
  static int buflen = 1024;
  int level = 0;
  int strlen = 0;

  if (!s)
    s = (char *)malloc (buflen);

  s[strlen++] = '(';
  
  while ((c = input()) != ')' || level != 0)
    {
      if (c == '(')
        level++;
      else if (c == ')')
        level--;
      else if (c == '\\')
        {
          s[strlen++] = c;
          c = input();
        }

      s[strlen++] = c;

      /* always keep space for at least 2 more chars so that we can
         append escaped chars without checking */
      if (strlen > buflen-2)
        {
          buflen += 1024;
          s = (char *)realloc (s, buflen);
        }
    }
  s[strlen++] = ')';
  s[strlen++] = '\0';

  return (s);
}

char *
read_binstring ()
{
  char c;
  static char *s = NULL;
  static int buflen = 1024;
  int strlen = 0;

  if (!s)
    s = (char *)malloc (buflen);
  
  s[strlen++] = '<';
  while ((c = input()) != '>')
    {
      if (c != '\n' && c != '\r')
        s[strlen++] = c;

      if (strlen > buflen - 2)
        {
          buflen += 1024;
          s = (char *)realloc (s, buflen);
        }
    }
  s[strlen++] = '>';
  s[strlen++] = '\0';

  return (s);
}  

int
check_com_arg ()
{
  Ps_op *p;
  if (p = find_ps_op (yytext))
    {
      yylval.ps_op = p;
      return (COMMAND);
    }
  else
    return (ARGUMENT);
}

char *
read_to_eol ()
{
  char c;
  static char *s = NULL;
  static int buflen = 1024;
  int strlen = 0;

  if (!s)
    s = (char *)malloc (buflen);
  
  while ((c = input()) != '\n' && c != '\r')
    {
      s[strlen++] = c;

      if (strlen > buflen-1)
        {
          buflen += 1024;
          s = (char *)realloc (s, buflen);
        }
    }
  s[strlen++] = '\0';
  
  return (s);
}
